#include <string>
#include <iostream>
#include <fstream>

#ifdef _MSC_VER
#pragma warning (push)
#pragma warning (disable : 4996)
#endif

#include <boost/iostreams/filtering_stream.hpp>
#include <boost/iostreams/filter/zlib.hpp>
#include <boost/iostreams/filter/gzip.hpp>
#include <boost/iostreams/filter/bzip2.hpp>
#include <boost/iostreams/device/file.hpp>
#include <boost/iostreams/copy.hpp>

#ifdef _MSC_VER
#pragma warning (pop)
#endif

#include <boost/filesystem.hpp>
#include <boost/progress.hpp>

const std::string PATH = "../data/";

enum Algorithm
{
	ZLIB = 0,
	GZIP,
	BZIP2
};

enum DataType
{
	RANDOM = 0,
	TEXT,
	EXECUTABLE
};

const std::string ALGORITHM_NAME[] =
{
	"zlib",
	"gz",
	"bz2"
};

const std::string FILE_NAME[] =
{
	"random",
	"text",
	"executable"
};

const std::string DOT = ".";

const std::string FILE_EXTENSION[] =
{
	"bin",
	"txt",
	"exe"
};

void PrintFileSize(const std::string& fileName)
{
	std::cout << boost::filesystem::file_size(fileName) << " bytes\t";
}

template <typename AlgorithmFunction>
void Compress(Algorithm algorithm, DataType type, AlgorithmFunction function)
{
	std::cout << ALGORITHM_NAME[algorithm] << "...\t\t\t";
	boost::progress_timer progress;

	const std::string sourceFileName =
		PATH + FILE_NAME[type] + DOT + FILE_EXTENSION[type];
	const std::string sinkFileName =
		PATH + FILE_NAME[type] + DOT + ALGORITHM_NAME[algorithm];

	boost::iostreams::filtering_ostreambuf out;
	out.push(function);
	out.push(boost::iostreams::file_sink(sinkFileName, std::ios::binary));
	boost::iostreams::copy(
		boost::iostreams::file_source(sourceFileName, std::ios::binary), out);

	PrintFileSize(sinkFileName);
}

template <typename AlgorithmFunction>
void Decompress(Algorithm algorithm, DataType type, AlgorithmFunction function)
{
	std::cout << ALGORITHM_NAME[algorithm] << "...\t\t\t\t\t";
	boost::progress_timer progress;

	const std::string sourceFileName =
		PATH + FILE_NAME[type] + DOT + ALGORITHM_NAME[algorithm];
	const std::string sinkFileName = 
		sourceFileName + DOT + FILE_EXTENSION[type];

	boost::iostreams::filtering_istreambuf in;
	in.push(function);
	in.push(boost::iostreams::file_source(sourceFileName, std::ios::binary));
	boost::iostreams::copy(
		in, boost::iostreams::file_sink(sinkFileName, std::ios::binary));
}

int main(int /*argc*/, char* /*argv*/[])
{
	boost::iostreams::zlib_params zlib_params;
	zlib_params.noheader = false;
	zlib_params.level = boost::iostreams::zlib::best_compression;

	boost::iostreams::gzip_params gzip_params;
	gzip_params.noheader = false;
	gzip_params.level = boost::iostreams::gzip::best_compression;

	boost::iostreams::zlib_compressor zlibCompressor(zlib_params);
	boost::iostreams::gzip_compressor gzipCompressor(gzip_params);
	boost::iostreams::bzip2_compressor bzip2Compressor;

	std::cout << "Compress files ===================================\n\n";

	std::cout << "Random binary file\t";
	PrintFileSize(PATH + FILE_NAME[RANDOM] + DOT + FILE_EXTENSION[RANDOM]);
	std::cout << " \n\n";
	Compress(ZLIB, RANDOM, zlibCompressor);
	Compress(GZIP, RANDOM, gzipCompressor);
	Compress(BZIP2, RANDOM, bzip2Compressor);

	std::cout << "Text file\t\t";
	PrintFileSize(PATH + FILE_NAME[TEXT] + DOT + FILE_EXTENSION[TEXT]);
	std::cout << " \n\n";
	Compress(ZLIB, TEXT, zlibCompressor);
	Compress(GZIP, TEXT, gzipCompressor);
	Compress(BZIP2, TEXT, bzip2Compressor);

	std::cout << "Executable file\t\t";
	PrintFileSize(PATH + FILE_NAME[EXECUTABLE] + DOT + FILE_EXTENSION[EXECUTABLE]);
	std::cout << " \n\n";
	Compress(ZLIB, EXECUTABLE, zlibCompressor);
	Compress(GZIP, EXECUTABLE, gzipCompressor);
	Compress(BZIP2, EXECUTABLE, bzip2Compressor);

	std::cout << "Decompress files =================================\n\n";

	boost::iostreams::zlib_decompressor zlibDecompressor(zlib_params);
	boost::iostreams::gzip_decompressor gzipDecompressor;
	boost::iostreams::bzip2_decompressor bzip2Decompressor;

	std::cout << "Random binary file\n\n";
	Decompress(ZLIB, RANDOM, zlibDecompressor);
	Decompress(GZIP, RANDOM, gzipDecompressor);
	Decompress(BZIP2, RANDOM, bzip2Decompressor);

	std::cout << "Text file\n\n";
	Decompress(ZLIB, TEXT, zlibDecompressor);
	Decompress(GZIP, TEXT, gzipDecompressor);
	Decompress(BZIP2, TEXT, bzip2Decompressor);

	std::cout << "Executable file\n\n";
	Decompress(ZLIB, EXECUTABLE, zlibDecompressor);
	Decompress(GZIP, EXECUTABLE, gzipDecompressor);
	Decompress(BZIP2, EXECUTABLE, bzip2Decompressor);

	return 0;
}
